import numpy as np
import pandas as pd

from .params import Params


def load_ffme_returns(cols=None):
    """
    Load the Fama-French Dataset for the returns of the Top and Bottom Deciles by MarketCap
    """
    if cols is None:
        cols = ["Lo 10", "Hi 10"]
    df = pd.read_csv(Params.DATA_FFME_PATH, index_col=0, na_values=-99.99)
    df.index = pd.to_datetime(df.index, format="%Y%m").to_period("M")
    df /= 100
    df = df[cols]
    df.rename({"Lo 10": "smallCap", "Hi 10": "largeCap"}, axis=1, inplace=True)
    return df


def load_hfi_returns():
    """
    Load and format the EDHEC Hedge Fund Index Returns
    """
    df = pd.read_csv(Params.DATA_HFI_PATH, index_col=0)
    df.index = pd.to_datetime(df.index).to_period("M")
    df /= 100
    return df


def dict_to_keys_values(dic):
    keys, values = list(dic.keys()), list(dic.values())
    return keys, values


def numpy_assert_almost_dicts(dict1, dict2):
    keys1, values1 = dict_to_keys_values(dict1)
    keys2, values2 = dict_to_keys_values(dict2)
    return np.testing.assert_equal(keys1, keys2), np.testing.assert_almost_equal(values1, values2)



import numpy as np
import pandas as pd
import pytest
from pal import utils


def test_load_ffme_returns():
    df = utils.load_ffme_returns()
    assert df.shape == (1110, 2)
    assert df.index[0] == pd.Period("1926-07")
    assert df.index[-1] == pd.Period("2018-12")
    assert df.columns.tolist() == ["smallCap", "largeCap"]


def test_load_hfi_returns():
    df = utils.load_hfi_returns()
    assert df.shape == (263, 13)
    assert df.index[0] == pd.Period("1997-01")
    assert df.index[-1] == pd.Period("2018-11")


def test_dict_to_keys_values():
    d = {"a": 1, "b": 2, "c": 3, "d": None}
    keys, values = utils.dict_to_keys_values(d)
    assert keys == ["a", "b", "c", "d"]
    assert values == [1, 2, 3, None]


def test_numpy_assert_almost_dicts():
    dict1 = {"b": np.array([1, 2, 0.2])}
    dict2 = {"b": np.array([1, 2, 3 * 0.1 - 0.1])}  # almost same
    dict3 = {"b": np.array([999, 888, 444])}  # completely different

    # no exception expected
    utils.numpy_assert_almost_dicts(dict1, dict2)

    # assert for expected failure
    with pytest.raises(AssertionError):
        utils.numpy_assert_almost_dicts(dict1, dict3)



import pandas as pd


def analyse_returns(returns: pd.Series, freq: int) -> dict:
    periods = len(returns)
    total_gross = (returns + 1).prod()
    total_ret = (returns + 1).prod() - 1
    per_period_gross = total_gross ** (1 / periods)
    per_period_ret = per_period_gross - 1
    annual_gross = per_period_gross ** freq
    annual_ret = annual_gross - 1
    return {
        "periods": periods,
        "total_gross": total_gross,
        "total_ret": total_ret,
        "per_period_gross": per_period_gross,
        "per_period_ret": per_period_ret,
        "annual_gross": annual_gross,
        "annual_ret": annual_ret,
    }


def analyse_risk(returns: pd.Series, freq: int) -> dict:
    std = returns.std()
    semi_dev = returns[returns < 0].std(ddof=0)
    var = returns.var()
    annual_vol = std * (freq ** 0.5)
    return {
        "std": std,
        "semi_dev": semi_dev,
        "var": var,
        "annual_vol": annual_vol,
    }


def calculate_annual_sharpe(annual_ret, annual_vol, risk_free_rate):
    return (annual_ret - risk_free_rate) / annual_vol


def simulate_wealth_index(returns, A):
    return A * (returns + 1).cumprod()


def get_prev_high(series: pd.Series):
    return series.cummax()


def get_prev_low(series: pd.Series):
    return series.cummin()


def calculate_max_drawdown(wealth_index, prev_high) -> dict:
    drawdown = (wealth_index - prev_high) / prev_high
    max_drawdown = drawdown.min()  # drawdown is negative
    id_max_drawdown = drawdown.idxmin()
    return {
        "wealth_index": wealth_index,
        "prev_high": prev_high,
        "drawdown": drawdown,
        "max_drawdown": max_drawdown,
        "id_max_drawdown": id_max_drawdown,
    }


def get_skewness(series: pd.Series):
    # normal skewness is 0
    num = ((series - series.mean()) ** 3).sum()
    denum = len(series) * series.std(ddof=0) ** 3
    return num / denum


def get_kurtosis(series: pd.Series, excess):
    # normal kurtosis is 3
    num = ((series - series.mean()) ** 4).sum()
    denum = len(series) * series.std(ddof=0) ** 4
    kurtosis = num / denum
    if excess:
        kurtosis -= 3
    return kurtosis


# def get_VaR_historic(r, var_pct: int = 95):
#     assert 1 <= var_pct <= 100
#     if isinstance(r, pd.DataFrame):
#         return r.aggregate(get_VaR_historic, var_pct=var_pct)
#     elif isinstance(r, pd.Series):
#         return -np.percentile(r, (100 - var_pct))
#     else:
#         raise TypeError("Expected DataFrame or Series")



import pandas as pd

import pal.analysis as analysis


class Portfolio:
    def __init__(self, returns: pd.Series, freq: int):
        """given returns, perform analysis

        Args:
            returns (Series-like): net returns in terms of percentage. e.g: 50% = 0.5
            freq (int): frequency the data is representing. e.g: 252 = daily, 12 = monthly, 52 = weekly
        """
        self.returns = returns
        self.freq = freq

        self.__analysed_returns = None
        self.__analysed_risk = None
        self.annual_sharpe = None
        self.wealth_index = None
        self.prev_high = None
        self.prev_low = None
        self.drawdown = None
        self.skewness = None
        self.kurtosis = None

    @property
    def analysed_returns(self):
        self.__analysed_returns = analysis.analyse_returns(returns=self.returns, freq=self.freq)
        return self.__analysed_returns

    @property
    def analysed_risk(self):
        self.__analysed_risk = analysis.analyse_risk(returns=self.returns, freq=self.freq)
        return self.__analysed_risk

    def get_annual_sharpe(self, risk_free_rate: float = 0):
        if not all([self.__analysed_returns, self.__analysed_risk]):
            _ = self.analysed_returns
            _ = self.analysed_risk
        self.annual_sharpe = analysis.calculate_annual_sharpe(
            annual_ret=self.__analysed_returns["annual_ret"],
            annual_vol=self.__analysed_risk["annual_vol"],
            risk_free_rate=risk_free_rate,
        )
        return self.annual_sharpe

    def get_wealth_index(self, A=1000):
        self.wealth_index = analysis.simulate_wealth_index(returns=self.returns, A=A)
        return self.wealth_index

    def get_prev_high(self, series: pd.Series):
        self.prev_high = analysis.get_prev_high(series=series)
        return self.prev_high

    def get_prev_low(self, series: pd.Series):
        self.prev_low = analysis.get_prev_low(series=series)
        return self.prev_low

    def get_max_drawdown(self, A=1000):
        if self.wealth_index is None or self.prev_high is None:
            self.get_wealth_index(A=A)
            self.get_prev_high(self.wealth_index)
        self.drawdown = analysis.calculate_max_drawdown(wealth_index=self.wealth_index, prev_high=self.prev_high)
        return self.drawdown

    def get_skewness(self):
        self.skewness = analysis.get_skewness(series=self.returns)
        return self.skewness

    def get_kurtosis(self, excess=False):
        # normal (non-excess) kurtosis is 3
        self.kurtosis = analysis.get_kurtosis(series=self.returns, excess=excess)
        return self.kurtosis



import numpy as np
from pal import analysis, utils

from .desired import Desired

ffme = utils.load_ffme_returns()
hfi = utils.load_hfi_returns()


def test_analyse_returns_annual_ret():
    desired = Desired.FFME_ANN_RET
    actual = analysis.analyse_returns(ffme, 12)["annual_ret"].to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_analyse_returns_per_period_ret():
    desired = Desired.FFME_PER_PERIOD_RET
    actual = analysis.analyse_returns(ffme, 12)["per_period_ret"].to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_analyse_risk_annual_vol():
    desired = Desired.FFME_ANN_VOL
    actual = analysis.analyse_risk(ffme, 12)["annual_vol"].to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_analyse_risk_semi_dev():
    desired = Desired.HFI_SEMI_DEV
    actual = analysis.analyse_risk(hfi, 12)["semi_dev"].to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_calculate_annual_sharpe():
    desired = Desired.FFME_ANN_SHARPE
    annual_ret = analysis.analyse_returns(ffme, 12)["annual_ret"]
    annual_vol = analysis.analyse_risk(ffme, 12)["annual_vol"]
    risk_free_rate = 0.0
    actual = analysis.calculate_annual_sharpe(
        annual_ret=annual_ret, annual_vol=annual_vol, risk_free_rate=risk_free_rate
    ).to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_calculate_annual_sharpe_3pct():
    desired = Desired.FFME_ANN_SHARPE_3PCT
    annual_ret = analysis.analyse_returns(ffme, 12)["annual_ret"]
    annual_vol = analysis.analyse_risk(ffme, 12)["annual_vol"]
    risk_free_rate = 0.03
    actual = analysis.calculate_annual_sharpe(
        annual_ret=annual_ret, annual_vol=annual_vol, risk_free_rate=risk_free_rate
    ).to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_simulate_wealth_index():
    desired = Desired.FFME_WEALTH_INDEX
    actual = analysis.simulate_wealth_index(ffme, 1000)["largeCap"].tolist()[:5]
    np.testing.assert_almost_equal(desired, actual)


def test_calculate_max_drawdown():
    desired = Desired.FFME_MAX_DRAWDOWN
    wealth_index = analysis.simulate_wealth_index(ffme, 1000)
    prev_high = analysis.get_prev_high(wealth_index)
    actual = analysis.calculate_max_drawdown(wealth_index=wealth_index, prev_high=prev_high)["max_drawdown"].to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_calculate_max_drawdown_index():
    desired = Desired.FFME_MAX_DRAWDOWN_PERIOD  # aka index
    wealth_index = analysis.simulate_wealth_index(ffme, 1000)
    prev_high = analysis.get_prev_high(wealth_index)
    actual = analysis.calculate_max_drawdown(wealth_index=wealth_index, prev_high=prev_high)[
        "id_max_drawdown"
    ].to_dict()
    np.testing.assert_equal(actual, desired)


def test_get_skewness():
    desired = Desired.HFI_SKEWNESS
    actual = analysis.get_skewness(hfi).sort_values().to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)


def test_get_kurtosis():
    desired = Desired.HFI_KURTOSIS
    actual = analysis.get_kurtosis(hfi, excess=False).to_dict()
    utils.numpy_assert_almost_dicts(desired, actual)



from setuptools import find_packages, setup

with open("README.md", "r") as file:
    long_description = file.read()

# # to-do: separate dev requirements
setup(
    name="portfolio-analysis-library",
    description="#WIP: Analysing Stock Returns & Constructing Portfolios with Python",
    version="0.0.3",
    author="Reza N",
    author_email="32329310+rezan21@users.noreply.github.com",
    url="https://github.com/rezan21/Portfolio-Analysis-Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests"]),
    python_requires=">=3.6",
    # py_modules=["analysis", "log", "main", "pal", "params", "utils"],
    # package_dir={"": "src"},
    install_requires=["numpy", "pandas", "matplotlib", "seaborn", "tqdm"],
    extras_require={"dev": ["pytest"]},
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
        "Topic :: Office/Business :: Financial :: Investment",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Financial and Insurance Industry",
    ],
)
